#include <qdos.h>

/********************************************************
 * Module to allow semaphores.							*
 ********************************************************/

extern int _NUM_SEM; /* Number of semaphores in all */

/* Structures needed */
struct job_wait {
	struct job_wait *jw_next; /* Next job waiting on this semaphore */
	long jw_jobid; /* Job id of this waiting job */
};

struct semaphore {
	char s_val; /* Semaphore itself */
	struct job_wait *s_jobs; /* Jobs waiting on this semaphore */
} **sems;

static char val; /* Local semaphore for monitor purposes */

/********************************************************
 * Routine to initialise semaphore routines.			*
 ********************************************************/

void initsems()
{
	register int i;

	sems = (struct semaphore **)getml( _NUM_SEM * sizeof( struct semaphore *));
	for( i = 0; i < _NUM_SEM; i++)
		sems[i] = NULL;
}

/********************************************************
 * Routine to create a semaphore.						*
 ********************************************************/

int creatsem()
{
	register int i;
	register struct semaphore *ps;

	if( num_thrd() )
		return -1; /* Can't do this if there are active threads */

	for( i =0; i < _NUM_SEM; i++) {
		if( sems[i] == NULL )
			break;
		if( sems[i]->s_val || sems[i]->s_jobs)
			return -1; /* More checks for active threads */
	}

	if( i == _NUM_SEM)
		return -1; /* No more semaphores */

	if((ps = sems[i] = (struct semaphore *)malloc(sizeof (struct semaphore )))==NULL)
		return -1;

	ps->s_val = 0;
	ps->s_jobs = NULL;
	return i;
}

/********************************************************
 * Routine to delete a semaphore.						*
 ********************************************************/

int killsem( sem )
int sem;
{
	register struct semaphore *ps = sems[sem];

	if( num_thrd() )
		return -1; /* Can't do this if there are active threads */

	if( ps == NULL)
		return -1;

	if( ps->s_val || ps->s_jobs )
		return -1;
	free( ps );
	sems[sem] = NULL;
	return 0;
}

/********************************************************
 * Routine to grab a semaphore.							*
 ********************************************************/

int getsem( sem )
int sem;
{
	register struct semaphore *ps = sems[sem];
	struct job_wait jw;

	if( ps == NULL)
		return -1;

	while( _psem( &val )) /* Wait untiil we have exclusive use of this code */
		; 

	if( _psem( &ps->s_val )) {
		/* Can't get the semaphore, suspend ourselves 'till we can */
		jw.jw_next = ps->s_jobs;
		jw.jw_jobid = getpid();
		/* Link ourselves onto the list */
		ps->s_jobs = &jw;
		val = 0; /* Release the lock on this code */
		mt_susjb( jw.jw_jobid, -1, NULL); /* Wait */
	}
	else
		val = 0; /* Just release the lock */
	return 0; /* We have now got the semaphore */
}

/********************************************************
 * Routine to release a semaphore, restarting the first	*
 * of any waiting jobs.									*
 ********************************************************/

int rlssem( sem )
int sem;
{
	register struct semaphore *ps = sems[sem];
	long sjobid, d1, d2, prior, d3;

	if( ps == NULL)
		return -1;

	while ( _psem( &val ))
		; /* Wait 'till we have exclusive use of this code */

	if( ps->s_jobs ) {
		/* There are jobs waiting on this semaphore */
		sjobid = ps->s_jobs->jw_jobid;
		prior = 1;
		while( prior > 0 ) {
			/* Ensure job we are to release has gone to sleep */
			d1 = d2 = sjobid;
			mt_jinf( &d1, &d2, &prior, &d3);
		}
		/* Remove it from linked list */
		ps->s_jobs = ps->s_jobs->jw_next;
		/* Job is now asleep, wake it up and return */
		mt_reljb( sjobid );
		val = 0; /* Release our lock */
		return 0;
	}

	/* There are no jobs waiting on this semaphore, zero it and exit */
	ps->s_val = 0;
	val = 0; /* Release lock */
	return 0;
}
